# Clonal classification accuracy
# Description: Computes the accuracy of clonal relationships from intMEMORI compared to ground truth
# Author: Alejandro Granados Castro
# Postdoctoral Scholar at Elowitz Lab
# California Institute of Technology
# agranado@caltech.edu
# Compiled: Oct-12-2020



# For a given ground truth lineage we want to :
# 1. reconstruct the tree using the sister-likelihood method
# 2. Evalute the clonal assignment compared to the ground truth

# Get clones from a single lineage
# The returned object is a table (array with names)
get.all.clones2020<-function(ground){

      genotypes = do.call(rbind, str_split(ground$tip.label, '_') )[,2]
      #table of frequencies for all genotypes
      a<-table(genotypes)
      #a[a>1]

    #      this_clone<-names(a[a>1])[7]
    #      this_size<-a[a>1][7]
    #      which_cells<-grep(this_clone , res[[8]]$tip.label)
    #Return the part of the table which includes the genotypes that appeared more than once
    #in this colony
      return(a[a>1])

}



# Main clonal score method:
# It calculates the clonal accuracy over all possible partitions of the tree (see details in Methods)
# uses Most Recent Common Ancestor to calculate sub-trees and scores
clonal.score2020<-function(id,memoirData = c() ,control=F,plot.all = F,pdf_file = "clonalPlot"){



      if(T){
        # tree is good quality then continue
        # inspect.tree returns a list where element 8 is the ground_truth tree
        ground_truth = read.newick(text = memoirData$ground[id] %>% as.character())
        reconstruction = read.newick(text = memoirData$rec[id] %>% as.character())

        # to calculate the random guess. negative CONTROL
        if(control)
          ground_truth$tip.label = sample(ground_truth$tip.label)




        dist_tips = distTips(ground_truth)
        dist_tips = as.matrix(dist_tips)
        barcodes = row.names(dist_tips)

        simple_barcodes= str_split(barcodes,"_",simplify=T)[,2]

        clones_here = get.all.clones2020(ground_truth)
        if(length(clones_here)>0){
          #for this barcodes, where are all the cells
          clone_score = c()
          best_Sb = c()
          best_Sa = c()
          clone_totaA = c()
          #for this BARCode
          for(i in 1:length(clones_here)){
              #i = 1
              # how many cells
              clone_size = length( which(simple_barcodes==names(clones_here[i])) )
              masked_barcodes = simple_barcodes
              masked_barcodes[ which(simple_barcodes==names(clones_here[i])) ] <- rep("A",clone_size)
              masked_barcodes[masked_barcodes != "A"] = "X"
              # > masked_barcodes
              # [1] "X" "X" "X" "X" "X" "X" "X" "A" "A" "A" "X" "X" "X"
              #rename ground truth
              ground_truth_masked = ground_truth
              ground_truth_masked$tip.label = masked_barcodes
              #get most common ancestor of this clade (includes ALL A)
              clone_mrca = MRCA( ground_truth ,which(simple_barcodes==names(clones_here[i])))
              # this is the largest tree that includes all A's
              clone_tree = extract.clade(ground_truth_masked, clone_mrca)
              # Here we need a function that iterates over all mrca partitions
              # Get all subtrees
              tree_partitions = subtrees(clone_tree)
              # Save score for all partitions
              partition_score = c()
              S_a_array = c()
              S_b_array = c()
              # Iterate over all partitions:
              Tot_A = clone_size # How many A's in the whole MRCA
              for(t in 1:length(tree_partitions)){
                  clone_subtree = tree_partitions[[t]] #this is a list
                  # how many A's in this partition
                    S_a = length(grep("A",clone_subtree$tip.label))
                  partition_score[t] = S_a / ( Tot_A + length(clone_subtree$tip.label) - S_a)
                   # S_a  /  Tot_A   +  cells_considered - S_a
                   # S_a  / Tot_A + S_b
                   # length(grep("A",clone_tree$tip.label)) / ( clone_size + length(clone_tree$tip.label) - length(grep("A",clone_tree$tip.label)) )
                   S_b = length(clone_subtree$tip.label) - S_a

                   # Save these numbers for false positive and false negative calculation
                   S_a_array[t] = S_a
                   S_b_array[t] = S_b


              }

              clone_score[i] = max(partition_score)
              # we can save the index of the best score and then get S_a, Tota_A, Sb
              best_score_index =which.max(partition_score)
              best_Sb[i] =  S_b_array[best_score_index]
              best_Sa[i] =  S_a_array[best_score_index]
              clone_totaA[i] = Tot_A
              #masked_barcodes = paste(as.character(1:length(masked_barcodes)),masked_barcodes,sep="" )
              #clone_score[i] = length(grep("A",clone_tree$tip.label)) / ( clone_size + length(clone_tree$tip.label) - length(grep("A",clone_tree$tip.label)) )
            }

          output_table = data.frame(totA  = clone_totaA, score =clone_score, Sa = best_Sa, Sb = best_Sb)
          row.names(output_table)<- names(clones_here)

          #Before returning the object, we will print the groung truth , reconstruction and output table to PDF
          if(plot.all ==T){
                pdf( paste(pdf_file, r[[1]],"_treePlot_RF_",toString(round(r[[4]],3)) ," _.pdf" ,sep = ""), height = 10, width = 17)

                  grid.arrange(ggdendrogram( as.hclust(force.ultrametric(r[[8]]) ), rotate = T )  ,
                      ggdendrogram(as.hclust(r[[9]]),rotate=T),tableGrob(output_table),
                            ncol = 3)


                  list_dend = dendlist( as.dendrogram(force.ultrametric(r[[8]])) , as.dendrogram(as.hclust(r[[9]])))
                  list_dend %>% untangle(method = "step1side") %>% # Find the best alignment layout
                  tanglegram(short=T,type="r",dLeaf_right = 0,dLeaf_left = -10,margin_inner = 7,highlight_branches_lwd = FALSE)                       # Draw the two dendrograms

                dev.off()

          }

          return(list(clone_totaA,clone_score, best_Sa, best_Sb))
        }else{
          return(NULL)
        }
      }else{
        return(NULL)
      }


}
