# Simulations of intMEMOIR recording
# Description: Simulates stochastic recording as cells divide
# Author: Alejandro Granados Castro
# Postdoctoral Scholar at Elowitz Lab
# California Institute of Technology
# agranado@caltech.edu
# Compiled: Oct-12-2020

# Simulation is restricted to toplogies that are observed experimentally


# Load dependencies
library(data.table)
library(dplyr)
library(ape)
library(data.tree)
library(stringr)
library(cluster)
library(phangorn)

# Set up the full path for the data
# Barcode data is in the folder FISH/
integrase_folder= "ReconstructionParameters/"
file.path=paste("../",integrase_folder,sep="")






# estimation of GLOBAL parameters:
#Experimentally measured edit rates per site

estim.params.global <-function(estimG = 3,file = "../ReconstructionParameters/editRate/allBarcodes.txt"){
  all.barcodes<-fread(file,head = F)
  rates<-apply(all.barcodes,2,table)
  rates.pct = rates/apply(rates,2,sum)
  #estim parameters per site:
  estimMu = 1-rates.pct[2,]^(1/estimG)
  #rates.pct[3,] means "X"
  estimAlpha = 1-rates.pct[3,]/(1-rates.pct[2,])

  return(list(estimMu,estimAlpha))
}


# GLOBAL PARAMETERS
params_global = estim.params.global(estimG = 4, fil = paste(file.path,"editRate/allBarcodes.txt",sep=""))
mu = params_global[[1]]
alpha = params_global[[2]]






# FISH processing functions
# These functions generate the necessary files for the reconstruction code to work
# The input files are the FISH data and the output is the main data.frame which we refer to as memoirData

# Output: memoirData which is a dataframe with the ground truth trees and some statistics
process.all.files <- function(files_to_process,remove_largeDels = F) {

  aa = lapply(files_to_process,barcode.processing,remove_largeDels)
  aa=do.call(rbind,lapply(aa,unlist))
  colnames(aa)<-c("fileID","fileName","nCells","edit","dels","ground")
  aa<-as.data.frame(aa)

  #data frame is in factor format
  #to make our lives easier we can convert it to numeric
  indx<-sapply(aa,is.factor)
  indx[2] = FALSE #this is character
  indx[6] = FALSE # this is character
  aa[indx] <- lapply(aa[indx], function(x) as.numeric(as.character(x)))

  return(aa)
}

barcode.processing <- function(file.idx,remove_largeDels = F){

        print( paste("processing file ", toString(file.idx), all.files$file.name[file.idx]))
        # Where to save the filtered barcode set
        barcode_path = paste(file.path,"filteredData/",sep="")

        file.name=all.files$file.name[file.idx]
        # file.path is global
        fileName=paste(file.path,"FISH/",file.name,sep="")

        # READ fish data from .txt
        posInfo = fread(paste(fileName,".txt",sep=""),sep="\t",colClasses = 'character')
        names(posInfo)<-c("cell","state")
        # some files have 3 as unknown state. We treat that as unedited:
        posInfo$state = str_replace_all(posInfo$state, c("3" = "1"))


        # This is the number of the cell
        posInfo$cell = as.character(as.numeric(substr(posInfo$cell,5,6)))

        barcodeLength =  nchar(as.character(posInfo$state[1]))

        # filter large deletions i.e. xxxxxxxxxx
        # state has the barcode readout
        # cell has the cell's name (number ID)
        keep_this_cells = !is.na(posInfo$cell) & !posInfo$state==paste(rep("0",barcodeLength),collapse="")

        # remove cells with large deletions: 4 or more deletions 0000
        if(remove_largeDels) keep_this_cells = keep_this_cells & !grepl("000",posInfo$state)
        # NOTE: I need to put some filter such that if we are left without cells after filtering, just skip this colony
        print( paste(toString(sum(keep_this_cells)), "cells included _  ", toString(sum(grepl("000",posInfo$state))) ," large deletions" ))
        #remove: IF (no cells ){ return (NULL ) } else { keep going}
        if(sum(keep_this_cells)>0){
                barcodes = posInfo$state[keep_this_cells]
                names(barcodes) =posInfo$cell[keep_this_cells]

                # WRITE the filtered barcodes to file.
                # This will match the new filtered ground truth for further reconstruction
                # WORKS
                write.table( data.frame(cell = names(barcodes), state=barcodes) , file=paste(file.path,'filteredData/', file.name,'.txt',sep=""), quote=FALSE, sep='\t',row.names = F)


                # Extract the ground truth from the file
                # Here the ground truth has ALL cells, including those with deletions, low-quality etc.
                ground.truth = all.files$newick[file.idx]

                true.tree=read.newick(text=toString(ground.truth))


                # barcodes is already a filtered list. We need to keep only those cells
                # so we filter everything else from the ground truth (dead cells, XXXXXX, etc. )
                alive.tree = filter.cells.groundTruth(barcodes,true.tree)

                #fraction of edited sites

                b=as.matrix(do.call(cbind,strsplit(barcodes,"")))
                d4=sum(b=="1")/prod(dim(b))
                # deletion rate:
                d0=sum(b=="0")/prod(dim(b))

                # FILE_ID   NCELLS    PR_EDIT   GROUND_TRUTH
                return(list(file.idx,file.name, length(barcodes), d4, d0,  write.tree(alive.tree)  ) )
                # WE can now save everything in a new data.frame
                # let.save individual components in
          }else{
            return(NULL)
          }

}

filter.cells.groudTruth <- function(barcodes,true.tree){

  # At this point barcodes is the list of cells that we want to keep for further analysis

  ground.labels = true.tree$tip.label
  # last recording frame
  # Cell ID has the last frame the sell was seen, so we keep the ones that were seen until the end of the experiment
  last.frame = 216 #we know this from the experiment
  lf.pattern = paste(toString(last.frame),"_",sep="")

  #last frame gives us alive cells while names(barcodes) gives us the cells that we filtered by diverse reasons
  keep.leaves = paste(last.frame,"_",names(barcodes),sep="")

  drop.leaves = ground.labels[-which( ground.labels %in% keep.leaves)]

  new.tree = drop.tip(true.tree,drop.leaves)

  return(new.tree)

}

# Run it many times to get a histogram / average dist for THIS ground_truth
pipelineSim<-function(ground_truth,mu,alpha,return.tree = F,n_barcodes = 10){


  if(n_barcodes>10){
      mu = rep(mu, ceiling(n_barcodes/10))
      alpha = rep(alpha,ceiling(n_barcodes/10))
  }
  initial_barcode = paste(rep("1",n_barcodes),collapse="")

  if(return.tree ==F){
    d = initSimulation(ground_truth,initial_barcode) %>%
          simulateGroundTruth(mu = mu, alpha = alpha) %>%
            convertToPhylo(ground_truth = ground_truth) %>%
              reconstructSimulation(mu = mu, alpha = alpha)
  }else{
    #return the simulated tree
    d = initSimulation(ground_truth,initial_barcode) %>%
          simulateGroundTruth(mu = mu, alpha = alpha) %>%
            convertToPhylo(ground_truth = ground_truth)

  }

  return(d)
}



# Main pipeline for simulation of recordin using the ground truth lineages
# To perform simulation we take the lineages directly from the memoirData data frame
# memoirData should include only trees that have been filtered by quality / number of cells

# if MEMO mode is T , then we remove duplicated barcodes and reconstruct the lineage of "clones"
# MEMO mode is the one reported on the manuscript
runAllMetrics<-function(id,memoirData = c(), n_repeats=50,n_barcodes=10,MEMO = F, rooted = F){
  print(paste( "processing tree", toString(id)))
  # in memoirData, lineages with 3 cells have score NA
  if(!MEMO){
    ground_truth = read.newick(text = as.character(memoirData$ground[id]))
  }else{
    ground_truth = read.newick(text = as.character(memoirData$MEMOtrees[id]))
  }

  # After converting the trees to MEMO mode, some are left with only 2 tips
  # We return NULL if the tree already was classified as NA or if they are <3 leaves after MEMO
  if(!is.na(memoirData$RF[id]) & length(ground_truth$tip.label)>3){


      if(n_barcodes>10){
          mu = rep(mu, ceiling(n_barcodes/10))
          alpha = rep(alpha,ceiling(n_barcodes/10))
      }

      d_membow = c()
      d_memoir = c()
      d_recursive = c()
      for(i in 1:n_repeats){
        this_node  =   initSimulation(ground_truth,initial_barcode = paste(rep("1",n_barcodes),collapse=""))
        ground_sim =   simulateGroundTruth(this_node,mu = mu, alpha = alpha)
        ground_phylo = convertToPhylo(ground_sim, ground_truth)
        manualTree = reconstructLineage(ground_phylo, mu,alpha,return_tree = T,clust.method="diana")

        d_memoir[i] = 1-RF.dist(ground_phylo,manualTree,normalize = T)

        #Try recursive reconstruction with the simulation data (update Sep 16th)
        manualTree_recursive = recursiveReconstruction(manualTree,mu = mu,alpha = alpha)
        d_recursive[i] =1- RF.dist( manualTree_recursive,ground_phylo,normalize = T)

        # New membow distance (no second round of reconstruction)
        d_membow[i] = reconstructMembow(ground_phylo,manualTree_recursive,mu, alpha, rooted = rooted)


      }
      return(list(d_membow,d_memoir,d_recursive , length(ground_phylo$tip.label)))



  }else{
    return(NULL)
  }


}

# Aux functions
# This function will get the barcode data from the simulated lineage
printBarcodes<-function(ground_sim){
  # Get all the leaves from the tree
  a<-ground_sim$leaves
  # For each leave, get the barcode
  sim_barcodes=c()
  for(i in 1:length(a)){
    sim_barcodes[i] = a[[i]]$barcode
  }
  return(sim_barcodes)
}

# Get the simulated lineage ready to be reconstructed
# returns ground_phylo (phylo object)
convertToPhylo<-function(ground_sim,ground_truth){

  # When we convert to phylo, the barcode information is lost
  # So we need to rename the leaves
  ground_phylo = as.phylo(ground_sim)
  # We extract leaves fromt he tree and save the barcodes as a list
  # The order is the same as in the phylo object
  sim_barcodes = printBarcodes(ground_sim)
  # So we can replace the names of the tips directly:
  ground_phylo$tip.label<-sim_barcodes

  cell_names = str_split(ground_truth$tip.label,"_",simplify=T)[,1]

  ground_phylo$tip.label<-paste(cell_names,ground_phylo$tip.label,sep="_")

  return(ground_phylo )

}

# Reconstruct a simulated lineage and return the RF score AND/OR the resulting tree
#returns RF dist for this simulated lineage
reconstructSimulation<-function(ground_phylo,mu,alpha,return_tree = F){

  #get the barcode and cell id (cell id is not neccessarily continuous numbers)
  barcodes = str_split(ground_phylo$tip.label,"_",simplify=T)[,2]
  cell_ids = str_split(ground_phylo$tip.label,"_",simplify=T)[,1]

  # translate the barcode data to the old notation uxr
  barcodes_urx = str_replace_all(barcodes, c("2" = "r", "1" = "u","0"="x"))
  barcodes_urx = str_replace_all(barcodes_urx, c("3" = "u"))

  #recontruct the tree
  matdist_=manualDistML_2(as.character(barcodes_urx),mu = mu ,alpha = alpha,nGen = 4 )

  row.names(matdist_)<- paste(cell_ids,barcodes,sep="_")
  colnames(matdist_)<- paste(cell_ids,barcodes,sep="_")

  hclust.tree = as.phylo(as.hclust( diana(as.dist(t(matdist_)))))

  #here ground_phylo is "the ground truth" which is a simulation but this is what we want
  d = 1- RF.dist(hclust.tree,ground_phylo,normalize =T)
  if(return_tree){
    return(hclust.tree)
  }else{
    return(d)
  }

}

# reconstructMembow is the main function that collapses identical barcodes into single leaves
# The parameter min.memo.tree sets the threshold on when to perform a second round of reconstruction
# Only when the a collapsed (MEMO) tree have less than min.memo.tree cells, we are going to reconstruct the barcodes again
# min.memo.tree = 3 (default) means NO second round of reconstruction
# For all figures in the manuscript, no second round of reconstruction was performed
reconstructMembow<-function(ground_phylo, manualTree,estimMu,estimAlpha,return_tree = F,
    min.memo.tree = 3,nGen = 4, rooted = F){

  unique.genotypes = unique(do.call(rbind,str_split(manualTree$tip.label,"_"))[,2])

  # check that tree has more that two unique leafs
  if(length(unique.genotypes)>2){

    new.tree = manualTree
    new.alive.tree = ground_phylo
        for(g in 1:length(unique.genotypes)){
            #drop.leaves = ground_phylo$tip.label[grep(unique.genotypes[g],ground_phylo$tip.label)]
            drop.leaves = manualTree$tip.label[grep(unique.genotypes[g],manualTree$tip.label)]
            #delete all except one

            if(length(drop.leaves)>1){
                for(i in 1:(length(drop.leaves)-1)){
                    #reconstruction
                    new.tree = drop.tip(new.tree,drop.leaves[i])
                    #ground truth:
                    new.alive.tree = drop.tip(new.alive.tree,drop.leaves[i])
                }
            }

        }

        # If there are only 3 unique genotypes there is no point in reconstructing the tree
        # since the RF distance is not well defined for n=3
        if(length(new.tree$tip.label)>3){
          # if after removing leaves we are left with only a few number of unique genotypes
          # we might want to re-reconstruct the barcodes using ward.D2 as DIANA does not work as well with too few cells
          if(length(new.alive.tree$tip.label)< min.memo.tree)
            new.tree = reconstructLineage(new.alive.tree,mu,alpha,return_tree=T,clust.method="ward.D2",nGen = nGen)

          this.score = 1-RF.dist(new.tree,new.alive.tree,normalize = T,rooted = rooted)
        }else{this.score = -1}
        #now we have trees with unique leaves
        # barcodes_ = new.tree$tip.label
        # barcodes.raw  =  substr(barcodes_,4,14)
        # barcodes.raw = str_replace_all(barcodes.raw, c("2" = "r", "1" = "u","0"="x"))
        #
        #
        # # using global parameters
        # matdist_2 = manualDistML_2(barcodes.raw,estimMu,estimAlpha,3)
        # colnames(matdist_2)<- substr(barcodes_,4,14); row.names(matdist_2)<- substr(barcodes_,4,14)
        # new.tree = as.phylo(as.hclust(diana(as.dist(t(matdist_2)))))
        #
        # new.alive.tree_ = new.alive.tree;
        # new.alive.tree_$tip.label = substr(new.alive.tree$tip.label,4,14)
        #
        # this.score = 1-RF.dist(new.alive.tree_,new.tree,normalize = T) #RF distance without number ID may20
        #
        # if(is.na(this.score)){
        #   #add a root to the tree, based on the initial state of the barcodes: 1111...1
        #   #then calcualte the distance using RF(x, root = T)
        #   this.score = 1-RF.dist(root(add.tips(new.tree, "1111111111", 4),"1111111111"),
        #       root(add.tips(new.alive.tree_, "1111111111", 4),"1111111111"),normalize = T,rooted = T)
        # }


  }else{this.score = -1; new.alive.tree = ground_phylo; new.tree = manualTree} #main IF end

  if(!return_tree){
    return(this.score)
  }else{
    return(list(new.alive.tree)) #,new.tree,this.score))
  }
}


# Negative control
randomControl <- function(id,memoirData, nrepeats = 10,MEMO = F){
  # This function will take a ground truth lineage given by the id
  # We are going to do random guess reconstruction which means just shuffling the leaves of the three
  # this, however assumes that we know the topology and we only want to put the barcodes in the right order.
  # Given that some topologies are easier to reconstruct than other, this assumption takes care of trivial reconstruction
  # since random guess will be probably as good as actual reconstruction

  # in memoirData, lineages with 3 cells have score NA
  if(!MEMO){
    ground_truth = read.newick(text = as.character(memoirData$ground[id]))
  }else{
    ground_truth = read.newick(text = as.character(memoirData$MEMOtrees[id]))
  }

  if(!is.na(memoirData$RF[id]) & length(ground_truth$tip.label)>3){



  # res = inspect.tree(id, return.tree = T,plot.all=F)
  # ground_truth = res[[8]]


    rand_dist=c()
    for(i in 1:nrepeats){
        ground_rand = ground_truth
        ground_rand$tip.label = sample(ground_rand$tip.label)
        rand_dist[i]=1-RF.dist(ground_truth,ground_rand,normalize=T)

    }
    return(rand_dist)

  }else{
    return(NULL)
  }

}



# Simulation Functions


# Functions for simulation of recording restricted to a given topology.
# we don't need to simulate the lineage but just follow the one that was observed in the experiment.
# This simulation uses a 10mer array (10 recording units) and simulates the edits based on experimentally-measured parameteres
# Parameters are estimated independently for each recording unit: each unit has its own edit rate, state frequencies


# Main function to simulate a recording event
recIntegrase_2019<-function(barcode,mu,alpha,type = "trit"){

  #barcodeLength=length(a)
  #fullBarcode = a
  #we want to edit only the portion that corresponds to
  #the currently active integrase

  a=strsplit(barcode,"")[[1]]


  for (c in 1:length(a)){
    #only mutate on unchanged elements
    if(a[c]=="1"){
      m_event<-runif(1,0,1) #random number
      if(m_event<mu[c]){
        #mutually exclusive events
        if(type=='trit'){
          trans_pr <-runif(1,0,1) #random number

          if(trans_pr<alpha[c]){
            a[c]="2"
          }else{
            a[c]="0"
          }
        }else if(type=='binary'){
          a[c]="0"
        }
      }
    }
  }

  fullBarcode= a

  return( paste(fullBarcode,sep="",collapse=""))


}


# From the data.tree object ready starting from a phylo object (ground truth)
# returns a Node object which is the root of the simulated lineage
initSimulation<-function(ground_truth, initial_barcode = "1111111111"){

  # name the internal nodes before converting
  ground_truth<-makeNodeLabel(ground_truth,prefix="")
  # convert to data.tree
  ground_node = as.Node(ground_truth)
  # check that things are correct then assign the barcode state to the first cell
  if(ground_node$isRoot)
    ground_node$barcode = initial_barcode

  return(ground_node)


}

# Main Function to simulate recording over an exisiting lineage (recursive)
# Input: a Node object from initSimulation()
simulateGroundTruth <- function(this_node,mu,alpha){

  if(this_node$isLeaf){
    return()
  }else{
    #set the new barcode by simulating a cell division recording
    this_node$children[[1]]$barcode = recIntegrase_2019(this_node$barcode,mu,alpha)
    this_node$children[[2]]$barcode = recIntegrase_2019(this_node$barcode,mu,alpha)

    # set the barcode as the name for the node
  #  this_node$children[[1]]$name =   this_node$children[[1]]$barcode
  #  this_node$children[[2]]$name =   this_node$children[[2]]$barcode


    simulateGroundTruth(this_node$children[[1]],mu,alpha)
    simulateGroundTruth(this_node$children[[2]],mu,alpha)

  }

  return(this_node)
}




# Entropy of barcodes states and reconstruction accuracy

# Calculation of shannon entropy per site
# Returns the vector for entropies for each site
# Main function for supplemntary Fig S8
barcodeEntropy<-function( ground_truth, matrix_entropy = F ){

  barcodes = str_split(ground_truth$tip.label,"_",simplify=T)[,2]
  barcode_matrix<-do.call(rbind,lapply(lapply(barcodes,str_split,"",simplify=T),as.numeric))
  h = c()

  if(matrix_entropy){
    return(entropy(barcode_matrix))
  }else{

    for(i in 1:dim(barcode_matrix)[2]){
        h[i]=entropy::entropy(table(barcode_matrix[,i])/dim(barcode_matrix)[1])
      }
    return(h)
  }
}

stringToPhylo <- function(x){ read.newick(text = x) }
editRatePhylo <- function(x){
  tip_list<-lapply(x$tip.label,str_split,'_')
  barcode_list = do.call(rbind,lapply(tip_list,unlist))[,2]
  barcode_matrix = do.call(rbind,lapply(barcode_list,str_split,'',simplify = T))
  return(1-sum(barcode_matrix =='1')/ length(barcode_matrix))
 }

#Make the main plot
makeEntropyPlots<-function(entropy_res,entro_threshold= 2.3){


  max_memoir_ecdf = ecdf(entropy_res$rec_score[entropy_res$entropy <entro_threshold & entropy_res$rec_score >=0])
  low_memoir_ecdf = ecdf(entropy_res$rec_score[entropy_res$entropy >=entro_threshold & entropy_res$rec_score >=0])
  all_memoir_ecdf = ecdf( entropy_res$rec_score[ entropy_res$rec_score>=0] )


  ax = seq(0,1,0.001)
  plot(ax,all_memoir_ecdf(ax),lwd = 3,type= 'l',col = 'gray',
        ylab = "Fraction of colonies",xlab='RF score',main="Barcode entropy predict reconstruction accuracy",
        cex = 1.2,cex.lab = 1.7, cex.axis = 1.3)
  lines(ax,low_memoir_ecdf(ax),lwd = 3,type= 'l',col = 'red')
  lines(ax,max_memoir_ecdf(ax),lwd = 3,type= 'l',col = 'blue')

  legend(0.01,0.9, legend = c('All','L MEMOIR','H MEMOIR'), fill = c('gray','red','blue' ))
}
